/*
   Copyright (C) 2009 Canonical, Ltd.
   Author: Kees Cook <kees@ubuntu.com>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, only version 3 of the License.

   This file is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with this program.  If not, see <http://www.gnu.org/licenses/>.

*/
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <assert.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <sys/capability.h>
#include <sys/prctl.h>

struct wrapper_command {
    char *name;
    char *cmdline;
    int   uses_uid;
    uid_t uid;
    int   uses_caps;
    cap_value_t caps;
};

#define LINE_MAX  1024
#define CMDS_CONF "/etc/eucalyptus/wrappers.conf"
#define CMDS_MAX  80
struct wrapper_command cmds[CMDS_MAX] = { };

void load_cmds(void)
{
    char buf[LINE_MAX];
    int count = 0;
    FILE *conf;

    if ( !(conf = fopen(CMDS_CONF,"r")) ) {
        perror(CMDS_CONF);
        exit(1);
    }
    while (fgets(buf, LINE_MAX-1, conf)) {
        char *step, *name, *cmdline, *capabilities;
        if (buf[0] == '#') continue;
        buf[strlen(buf)] = '\n';
        step = buf;
        name = strsep(&step, " \t\n");
        if (!name || strlen(name) <= 0) continue;
        while (*step == ' ' || *step == '\t' || *step == '\n') step++;

        cmdline = strsep(&step, " \t\n");
        if (!cmdline || strlen(cmdline) <= 0) continue;
        while (*step == ' ' || *step == '\t' || *step == '\n') step++;

        capabilities = strsep(&step, " \t\n");
        if (!capabilities || strlen(capabilities) <= 0) continue;

        // printf("NAME: |%s| CMD: |%s| CAP: |%s|\n", name, cmdline, capabilities);

        assert (count<CMDS_MAX);
        cmds[count].name = strdup(name);
        cmds[count].cmdline = strdup(cmdline);
        if (capabilities[0] != '\0') {
            if (strncmp(capabilities,"cap",3)==0) {
                if (cap_from_name(capabilities, &cmds[count].caps)) {
                    fprintf(stderr,"Bad capabilities\n");
                    exit(1);
                }
                cmds[count].uses_caps = 1;
            }
            else {
                if (isdigit(*capabilities)) {
                    cmds[count].uses_uid = 1;
                    cmds[count].uid = atoi(capabilities);
                }
            }
        }
        count++;
    }
    fclose(conf);
}

#if 0
void dump(void)
{
    struct wrapper_command *cmd;
    for (cmd=cmds; cmd && cmd->name; cmd++) {
        printf("[%s] [%s] ", cmd->name, cmd->cmdline);
        if (cmd->uses_uid)
            printf("uid:%d", cmd->uid);
        else if (cmd->uses_caps) {
            char *cap_string = cap_to_name(cmd->caps);
            printf("cap:%s(%d)", cap_string, cmd->caps);
            cap_free(cap_string);
        }
        printf("\n");
    }
}
#endif

void run(struct wrapper_command *cmd, int argc, char *argv[])
{
    uid_t uid = getuid();
    gid_t gid = getgid();
    struct passwd *entry=NULL;
    int i=0;
    char **args=NULL;
    cap_t caps;

    if (cmd->uses_uid) {
        uid = cmd->uid;
    }
    if (cmd->uses_caps) {
        /* Set capabilities */

        /* Unfortunately, capabilities are unconditionally cleared across exec().
           So this wrapper is a bit useless at the moment in its attempt to limit
           privileges.
            http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=439873
            http://lkml.indiana.edu/hypermail/linux/kernel/0704.2/0667.html
         */

        assert (caps = cap_init());
        assert (cap_clear(caps) == 0);
        assert (cap_set_flag(caps, CAP_EFFECTIVE,   1, &cmd->caps, CAP_SET) == 0);
        assert (cap_set_flag(caps, CAP_PERMITTED,   1, &cmd->caps, CAP_SET) == 0);
        assert (cap_set_flag(caps, CAP_INHERITABLE, 1, &cmd->caps, CAP_SET) == 0);
        assert (cap_set_proc(caps) == 0);

        assert (prctl(PR_SET_KEEPCAPS, 1, 0, 0, 0) == 0);
    }

    if ( !(entry = getpwuid(uid)) ) {
        perror("getpwuid");
        exit(253);
    }

    uid = entry->pw_uid;
    gid = entry->pw_gid;

    /* If we're changing user id, drop all the supplemental groups */
    if (uid != getuid()) {
        if (setgroups(0, NULL)) {
            perror("setgroups");
            exit(253);
        }
    }

    /* Set group */
    if (setresgid(gid, gid, gid)) {
        perror("setresgid");
        exit(253);
    }
    /* Set uid */
    if (setresuid(uid, uid, uid)) {
        perror("setresuid");
        exit(253);
    }

    if (cmd->uses_caps) {
        /* Re-gain capabilities */
        assert (cap_set_proc(caps) == 0);
        assert (cap_free(caps) == 0);
    }

    if ( !(args = calloc(argc, sizeof(char*))) ) {
        perror("calloc");
        exit(253);
    }

    args[0] = strdup(cmd->cmdline);
    for (i = 2; i<argc; i++) {
        args[i-1] = strdup(argv[i]);
    }
    execv(args[0], args);
    perror(args[0]);
}

int main(int argc, char *argv[])
{
    struct wrapper_command *cmd=NULL;
    char *cleancmd=NULL;

    if (argc>1) {
        load_cmds();

        cleancmd = strdup(argv[1]);
        cleancmd = basename(cleancmd);

        for (cmd=cmds; cmd && cmd->name; cmd++) {
            if (!cleancmd) continue;
            if (strcmp(cleancmd,cmd->name) == 0 ||
                strcmp(argv[1],cmd->cmdline) == 0) {

                run(cmd, argc, argv);
                exit(255);
            }
        }
        fprintf(stderr,"ERROR: Disallowed command %s\n", argv[1]);
    }
    else {
        fprintf(stderr, "ERROR: must supply a command to run, as a command-line argument.\n");
    }
    exit(127);
}
